// Entry point for the interface compiler
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.
//
// The idlc program implements two main functions.  One is to turn
// .idl files into binary type descriptions.  The other is to turn
// those type descriptions into language-specific stubs (though a
// compiler/interpreter could also just read the binary types
// directly, as the ORB does).
//
// For the former function, it is necessary to do the work in multiple
// passes, as not all information is available the first time through due
// to the absence of forward declarations.  The passes are as follows:
//
//   1: Parse the input text, and determine which symbols exist in
//      which namespaces.  At this point, no non-namespace-qualified
//      symbol lookups will be done (though namespace-qualified lookups
//      are done, such as to check for duplicate symbols).
//
//      Objects will be created for each entity (Interface, Struct,
//      Datum, Alias, etc.), but any reference to another entity will be
//      stored as a StrList, not as an object pointer.  This must be the
//      case even if a lookup during this pass would have found
//      something, as on the second pass it might find a "closer"
//      matching symbol added later.
//
//      The StrList will later be looked up with only the entity
//      containing the reference as context, so namespace search rules
//      must not depend on anything else.
//
//      This is the only pass that involves actually reading and
//      parsing the input file(s).
//
//   2: Look up the namespace names of using namespace.* statements.
//      These are the only lookups where order of the statements in the
//      file matters, and it alters the result of a lookup, so it must be
//      completed before chain lookups are done.
//
//   3: Look up the symbol names recorded in pass 1 for aliases, and
//      typedefs, and const datum initializers and types. This is done
//      before the rest of the lookups, so there will always be a
//      complete chain to follow in subsequent passes.
//
//      This is done by calling the virtual lookup_chain() method on
//      the root namespace, which recursively calls it on all of its
//      contents.
//
//   4: Look up other symbol names recorded in pass 1.  At this point,
//      unresolved alias chains may still exist and must be handled,
//      though they should be fixed during this phase so that they do not
//      appear in phase 4.  Typedefs will remain fully chained, as the
//      chain is semantically meaningful and must be preserved in the
//      output. 
//      
//      Const datum initializers are looked up in this pass despite being
//      chains, since aliases aren't resolved in pass 2, and it's
//      (slightly) easier to do the lookup in pass 3 and resolve the
//      chain in pass 4 than to store a Symbol reference in pass 2,
//      get the concerete sym in pass 3, and still have to either
//      concrete-ize other Datums' initializers or wait until pass 4
//      to resolve the chain.
//
//      This is done by calling the virtual lookup_misc() method on the
//      root namespace, which recursively calls it on all of its
//      contents.  Upon receiving this call, each entity object should
//      call lookup_sym() or lookup_type() on the various StrLists it has
//      stored, but should not do anything with the object pointer
//      until pass 4 (unless the object was satisfactorily initialized
//      in pass 1).
//
//   5: Perform any remaining semantic analysis, now that all references
//      to other entities have been looked up.  
//
//      This is done by calling the virtual final_analysis() method on
//      the root namespace, which recursively calls it on all of its
//      contents.
//
//   6: Generate output.
//      This is done by calling the virtual output() method on
//      the root namespace, which recursively calls it on all of its
//      contents.
//
// All circular dependencies must be contained within the set of files
// passed to the compiler in one invocation.  Supporting circular
// dependencies over multiple invocations of the compiler would require
// explicitly running only certain passes, then compiling the other
// batch, and then finishing the original batch.  It could be done with
// some pain (intermediate states would need to be serialized), but I
// only want to do it if there's a real need.  Circular dependencies
// ought to be isolated fairly locally...

#include <climits>
#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <cstdlib>
#include <cerrno>

#include <idlc.h>
#include <lang.h>
#include <targets.h>
#include <util.h>

int num_err;
unsigned int enum_pos;
bool no_output;
bool server_stubs;
bool makedep;
int current_pass;
int traversal;
Con *yylval_con;
String **yylval_string;

UserNameSpaceRef toplevel, cdl_toplevel;
UserNameSpace *output_ns;
NameSpaceRef cur_nspace;
const char *cmdname, *output_dir, *cur_input_file = "<none>", *output_ns_name;

list<const char *> inputs;
list<NameSpaceRef> nspace_stack;
Language *first_lang, *output_lang;
InterfaceRef System_Object;
StructRef System_VStruct;
AutoReleasePool autorelease_pool;

void print_usage()
{
	fprintf(stderr,
"Usage: %s [options] [<input(s)>]\n\n"
"Options are:\n"
"-d (--debug)           Print parser debugging information.\n"
"-h (--help)            Print this help screen.\n"
"-i DIR (--include=DIR) Look in DIR for resolving external type definitions.\n"
"                       Multiple include directories can be specified, and\n"
"                       will be mounted at the --namespace specified when\n"
"                       they were created.\n"
"-l LANG (--lang=LANG)  Generate stubs for this language.\n"
"-M (--makedep)         Output dependency information for make.\n"
"-n (--no-output)       Do not produce output; only check for errors.\n"
"-o DIR (--output=DIR)  Generate output in DIR.\n"
"-r (--server)          Generate server stubs according to the input CDL files.\n"
"-s NS (--namespace=NS) Input files constitute the given namespace.\n"
"-t TARGET (--target=T) Generate stubs for the specified target architecture.\n"
"                       By default, the current architecture is targeted.\n"
"--show-languages       List languages for which stub generation is supported.\n"
"--show-targets         List supported target architectures.\n"
"If a language is specified, language-specific stubs for the output --namespace\n"
"will be placed in the specified output directory.  All inputs must be specified\n"
"as --includes (including all components of the output namespace and anything\n"
"referenced by them).  No non-option inputs may be specified.\n\n"
"If a language is not specified, inputs are treated as IDL source\n"
"files, and are compiled into the specified output directory, as\n"
"the specified output --namespace.\n\n",
cmdname);
}

static void set_lang(const char *lang)
{
	if (output_lang) {
		fprintf(stderr, "Only one language may be specified.\n");
		throw UserError();
	}

	for (Language *l = first_lang; l; l = l->next) {
		if (!strcasecmp(l->name, lang)) {
			output_lang = l;
			return;
		}
	}
	
	fprintf(stderr, "Unsupported language \"%s\".\n", lang);
	throw UserError();
}

static void set_target(const char *targ)
{
	if (target != &targets[0]) {
		fprintf(stderr, "Only one target architecture may be specified.\n");
		throw UserError();
	}
	
	for (int i = 1; i <= max_target; i++) {
		if (!strcasecmp(targ, targets[i].name)) {
			target = &targets[i];
			return;
		}
	}

	fprintf(stderr, "Unsupported target \"%s\".\n", targ);
	throw UserError();
}

// Returns true if the argument is consumed.
static bool shortopt(char c, const char *arg)
{
	switch (c) {
		case 'd':
			idl_debug = 1;
			cdl_debug = 1;
			break;

		case 'h':
			print_usage();
			exit(0);
			
		case 'i':
			if (!arg) {
				fprintf(stderr, "The -i option requires an argument.\n");
				throw UserError();
			}
			
			UserNameSpace::declare_import(arg);
			return true;
		
		case 'M':
			makedep = true;
			break;
		
		case 'n':
			no_output = true;
			break;
		
		case 'l':
			if (!arg) {
				fprintf(stderr, "The -l option requires an argument.\n");
				throw UserError();
			}

			set_lang(arg);
			return true;
		
		case 'o':
			if (output_dir) {
				fprintf(stderr, "Only one output directory may be specified.\n");
				throw UserError();
			}
		
			if (!arg) {
				fprintf(stderr, "The -o option requires an argument.\n");
				throw UserError();
			}

			output_dir = arg;
			return true;
		
		case 'r':
			server_stubs = true;
			break;

		case 's':
			if (output_ns_name) {
				fprintf(stderr, "Only one output namespace may be specified.\n");
				throw UserError();
			}
		
			output_ns_name = arg;
			return true;

		case 't':
			if (!arg) {
				fprintf(stderr, "The -t option requires an argument.\n");
				throw UserError();
			}
			
			set_target(arg);
			return true;

		default:
			fprintf(stderr, "%s: unknown option %c\n", cmdname, c);
			throw UserError();
	}
	
	return false;
}

static void longopt(const char *s, const char *arg)
{
	if (!strcmp(s, "help")) {
		print_usage();
		exit(0);
	} else if (!strcmp(s, "debug")) {
		idl_debug = 1;
		cdl_debug = 1;
	} else if (!strcmp(s, "include")) {
		if (!arg) {
			fprintf(stderr, "The --include option requires an argument.\n");
			throw UserError();
		}

		UserNameSpace::declare_import(arg);
	} else if (!strcmp(s, "makedep")) {
		makedep = true;
	} else if (!strcmp(s, "no-output")) {
		no_output = true;
	} else if (!strcmp(s, "output")) {
		if (output_dir) {
			fprintf(stderr, "Only one output directory may be specified.\n");
			throw UserError();
		}

		if (!arg) {
			fprintf(stderr, "The --output option requires an argument.\n");
			throw UserError();
		}

		output_dir = arg;
	} else if (!strcmp(s, "namespace")) {
		if (output_ns_name) {
			fprintf(stderr, "Only one output namespace may be specified.\n");
			throw UserError();
		}

		if (!arg) {
			fprintf(stderr, "The --namespace option requires an argument.\n");
			throw UserError();
		}
	
		output_ns_name = arg;
	} else if (!strcmp(s, "language")) {
		if (!arg) {
			fprintf(stderr, "The --language option requires an argument.\n");
			throw UserError();
		}

		set_lang(arg);
	} else if (!strcmp(s, "target")) {
		if (!arg) {
			fprintf(stderr, "The --target option requires an argument.\n");
			throw UserError();
		}

		set_target(arg);
	} else if (!strcmp(s, "show-languages")) {
		printf("Supported language bindings:\n");
		for (Language *l = first_lang; l; l = l->next)	
			printf("   %s\n", l->name);
		printf("\n");
		exit(0);
	} else if (!strcmp(s, "show-targets")) {
		printf("Supported target architectures:\n");
		for (int i = 1; i <= max_target; i++)
			printf("   %s\n", targets[i].name);
		printf("\n");
		exit(0);
	} else if (!strcmp(s, "server")) {
		server_stubs = true;
	} else {
		fprintf(stderr, "%s: unknown option \"%s\"\n", cmdname, s);
		throw UserError();
	}
}

static int global_argc;
static const char **global_argv;
static int got_dashdash;

static void process_args(void)
{
	int i;
	size_t j;
	char *s;
	
	for (i = 1; i < global_argc; i++) {
		if (global_argv[i][0] == '-' && global_argv[i][1] && !got_dashdash) {
			const char *opt = global_argv[i];
		
			if (opt[1] == '-') {
				if (!opt[2]) {
					got_dashdash = 1;
				} else {
					s = strchr(&opt[2], '=');

					if (!s)
						longopt(&opt[2], NULL);
					else {
						*s=0;
						longopt(&opt[2], s+1);
					}
				}
			} else for (j = 1; j < strlen(opt); j++) {
				if (shortopt(opt[j], global_argv[i + 1]))
					i++;
			}
		} else {
			inputs.push_back(global_argv[i]);
		}
	}
}

// Provide automatic closing of a file handle upon exception
struct FileHolder {
	FILE *f;
	
	FileHolder()
	{
		f = NULL;
	}
	
	~FileHolder()
	{
		if (f)
			fclose(f);
	}
};

void parse_inputs(bool cdl)
{
	for (list<const char *>::iterator i = inputs.begin();
	     i != inputs.end(); ++i)
	{
		FileHolder fh;
		cur_input_file = *i;

		curline = 1;
		
		if (strcmp(cur_input_file, "-")) {
			yyin = fh.f = fopen(cur_input_file, "r");
			if (!yyin) {
				fprintf(stderr, "Cannot open input file \"%s\": %s\n",
				        cur_input_file, strerror(errno));
				throw UserError();
			}
		} else {
			yyin = stdin;
			cur_input_file = "stdin";
		}
		
		if (cdl) {
			cur_nspace = cdl_toplevel;
			setup_cdlparse();
			
			if (cdl_parse())
				throw UserError();
		} else {
			cur_nspace = output_ns;
			setup_idlparse();

			if (idl_parse())
				throw UserError();
		}
		
		if (finish_lex() || num_err)
			throw UserError();
	}
	
	cur_input_file = "<none>";
	curline = 0;

	if (!nspace_stack.empty())
		BUG();
}

static void init_output_ns()
{
	StrList *strl;

	try {
		String *str = new String(output_ns_name);
		strl = new StrList(str);
	}

	catch (InvalidArgument) {
		fprintf(stderr, "Output namespace \"%s\" is not vaild.\n",
		        output_ns_name);
		throw UserError();
	}

	cur_nspace = toplevel;
	NameSpace *new_ns = add_nspace(strl, false);

	if (!new_ns)
		BUG();
		
	NameSpace *conflict = check_for_imports(new_ns);
	if (conflict) {		
		fprintf(stderr, "Output namespace \"%s\" conflicts"
		                " with \"%s\" at \"%s\".\n",
		        output_ns_name, conflict->get_fq_name()->flatten()->c_str(),
		        conflict->get_path()->c_str());

		throw UserError();
	}

	if (!new_ns)
		BUG();
	output_ns = dynamic_cast<UserNameSpace *>(new_ns);
	if (!output_ns)
		BUG();
}

void lookup_system_object()
{
	try {
		StrList *sl = new StrList;
		
		sl->push_back(new String(""));
		sl->push_back(new String("System"));
		sl->push_back(new String("Object"));

		Symbol *sym = lookup_sym(toplevel, sl, toplevel);
		System_Object = dynamic_cast<Interface *>(sym);
	}
	
	catch (UserError) {
		yyerrorf("Could not find System.Object.");
		yyerrorf("Be sure to include the file or directory containing "
		         "the System namespace.");

		throw UserError();
	}
	
	if (!System_Object) {
		yyerrorf("System.Object is not an interface.");
		throw UserError();
	}
}

void lookup_system_vstruct()
{
	try {
		StrList *sl = new StrList();
		
		sl->push_back(new String(""));
		sl->push_back(new String("System"));
		sl->push_back(new String("VStruct"));

		Symbol *sym = lookup_sym(toplevel, sl, toplevel);
		System_VStruct = dynamic_cast<Struct *>(sym);
	}
	
	catch (UserError) {
		yyerrorf("Could not find System.VStruct.");
		yyerrorf("Be sure to include the file or directory containing "
		         "the System namespace.");

		throw UserError();
	}
	
	if (!System_VStruct) {
		yyerrorf("System.VStruct is not a struct.");
		throw UserError();
	}
	
	if (!System_VStruct->is_virtual()) {
		yyerrorf("System.VStruct is not virtual.");
		throw UserError();
	}
}

void normal_output()
{
	if (inputs.empty()) {
		fprintf(stderr, "No inputs given.\n");
		print_usage();
		throw UserError();
	}
	
	if (makedep) {
		fprintf(stderr, "Dependency generation is only supported "
		                "when generating language output.\n");
		throw UserError();
	}
	
	if (!no_output && !output_dir) {
		fprintf(stderr, "An output directory must be specified "
		                "when not using -n.\n");
		throw UserError();
	}

	if (!output_ns_name) {
		fprintf(stderr, "An output namespace must be specified.\n");
		throw UserError();
	}

	if (!no_output && output_dir)
		makepath(output_dir);

	init_output_ns();
	
	current_pass = 1;
	autorelease_pool.clean();

	parse_inputs(false);

	current_pass = 2;
	autorelease_pool.clean();

	lookup_system_object();
	lookup_system_vstruct();
	output_ns->lookup_imports();
	
	current_pass = 3;
	autorelease_pool.clean();
	
	output_ns->lookup_chain();
	if (num_err)
		throw UserError();
	
	current_pass = 4;
	autorelease_pool.clean();

	output_ns->lookup_misc();
	if (num_err)
		throw UserError();

	current_pass = 5;
	autorelease_pool.clean();

	output_ns->final_analysis();
	if (num_err)
		throw UserError();

	current_pass = 6;
	autorelease_pool.clean();
	
	if (!no_output)
		output_ns->output(output_dir);

	autorelease_pool.clean();
}

void language_output()
{
	if (server_stubs) {
		if (inputs.empty()) {
			fprintf(stderr, "No inputs given.\n");
			print_usage();
			throw UserError();
		}
	} else {
		if (!inputs.empty()) {
			fprintf(stderr, "Inputs may not be specified when generating "
			                "language-specific client output.\n"
			                "Use --include to specify input namespaces.\n");
			throw UserError();
		}

		if (!output_ns_name) {
			fprintf(stderr, "An output namespace must be specified.\n");
			throw UserError();
		}

		if (makedep) {
			fprintf(stderr, "Dependency information is currently only supported "
			                "for server stubs.\n");
			throw UserError();
		}
	}

	if (no_output) {
		fprintf(stderr, "-n may not be used when generating "
		                "language-specific output.\n");
		throw UserError();
	}

	if (!output_dir) {
		fprintf(stderr, "An output directory must be specified when "
		                "generating language-specific output.\n");
		throw UserError();
	}

	if (!makedep && target == &targets[0]) {
		fprintf(stderr, "Target detection not yet implemented.\n"
		                "Please specify a target architecture.\n");
			throw UserError();
	}

	if (!makedep)
		makepath(output_dir);

	if (server_stubs) {
		current_pass = INT_MAX;
		lookup_system_object();
		lookup_system_vstruct();
		parse_inputs(true);
		autorelease_pool.clean();
		output_lang->output_server(cdl_toplevel, output_dir);
	} else {
		StrList *output_ns_strl;
		try {
			String *output_ns_str = new String(output_ns_name);
			output_ns_strl = new StrList(output_ns_str);
		}
		
		catch (InvalidArgument) {
			fprintf(stderr, "Output namespace \"%s\" is not vaild.\n",
			        output_ns_name);
			throw UserError();
		}
		
		cur_nspace = toplevel;
		current_pass = INT_MAX;
		lookup_system_object();
		lookup_system_vstruct();
		
		Symbol *sym = lookup_sym(toplevel, output_ns_strl, toplevel);
		output_ns = dynamic_cast<UserNameSpace *>(sym);
		if (!output_ns) {
			fprintf(stderr, "Output namespace \"%s\" is not a pure namespace.\n",
			        output_ns_name);
			throw UserError();
		}
		
		output_ns->import_all_recursive();
		autorelease_pool.clean();
		output_lang->output_root(output_ns, output_dir);
	}

	autorelease_pool.clean();
}

int run_idlc(int argc, const char **argv)
{
	global_argc = argc;
	global_argv = argv;
	cmdname = argv[0];

	toplevel = new UserNameSpace();
	cdl_toplevel = new UserNameSpace();

	try {
		process_args();
			
		if (output_lang)
			language_output();
		else {
			if (server_stubs) {
				fprintf(stderr, "The -r (--server) option may only be used "
				                "with the -l (--lang) option.\n");

				throw UserError();
			}
		
			normal_output();
		}
	}
	
	catch (InternalError &e) {
		fprintf(stderr, "Internal idlc error at %s:%d\n",
		        e.file, e.line);
		return 1;
	}
	
	catch (InvalidArgument) {
		fprintf(stderr, "Internal idlc error: Uncaught InvalidArgument\n");
		return 1;
	}

	catch (SymbolNotFound) {
		fprintf(stderr, "Internal idlc error: Uncaught SymbolNotFound\n");
		return 1;
	}

	catch (DuplicateSymbol) {
		fprintf(stderr, "Internal idlc error: Uncaught DuplicateSymbol\n");
		return 1;
	}
	
	catch (UserError) {
		// An error message has already been displayed.
		num_err++;
	}
	
	catch (...) {
		fprintf(stderr, "%s:%d: Internal error: Uncaught Exception\n",
		        cur_input_file, curline);
		return 1;
	}

	if (num_err)
		return 1;

	return 0;
}

int main(int argc, const char **argv)
{
	int ret = run_idlc(argc, argv);
	autorelease_pool.clean();

	return ret;
}

extern "C" int yywrap()
{
	return 1;
}

extern char *yytext;

void idl_error(const char *s)
{
	if (strlen(yytext))
		fprintf(stderr, "%s:%d: %s at \"%s\".\n", cur_input_file, curline, s, yytext);
	else
		fprintf(stderr, "%s:%d: %s at end of input.\n", cur_input_file, curline, s);

	num_err++;
}

void yyerrorf(const char *s, ...)
{
	va_list va;
	va_start(va, s);
	fprintf(stderr, "%s:%d: ", cur_input_file, curline);
	vfprintf(stderr, s, va);
	fprintf(stderr, "\n");
	va_end(va);
	
	num_err++;
}

void yyerrorfl(const char *file, int line, const char *s, ...)
{
	va_list va;
	va_start(va, s);
	fprintf(stderr, "%s:%d: ", file, line);
	vfprintf(stderr, s, va);
	fprintf(stderr, "\n");
	va_end(va);
	
	num_err++;
}

int idl_lex()
{
	return yylex();
}
