// arch/x64/misc.cc -- Misc. arch-specific stuff
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/libc.h>
#include <kern/arch.h>
#include <kern/i8259.h>
#include <kern/time.h>
#include <kern/thread.h>
#include <kern/mem.h>

#include <arch/addrs.h>
#include <arch/multiboot.h>

extern u64 x64_init_ptbl_l4[512];

struct X64InitStack {
	u8 stack[4096 - ::Threads::thread_size];
	::Threads::Thread thread;
} __attribute__((aligned(4096))) x64_init_stack;

namespace Arch {
	namespace Priv {
		void set_idt();
		
		void show_regs(u64 *stack) {
			for (int i = 0; i < 16; i += 4)
				printf("r%02d 0x%016llx 0x%016llx 0x%016llx 0x%016llx\n",
				       i, stack[i], stack[i + 1], stack[i + 2], stack[i + 3]);
			
			printf("orig rsp: 0x%016llx rflags: 0x%016llx\n",
			       stack[19], stack[18]);
			
			printf("Thread %p (%s)\n", curthread, curthread->name);
			
			printf("Stack trace:       ");
			u64 *frame = (u64 *)stack[5];
			
			for (int i = 1; i < 32; i++) {
				u64 stackptr = frame[1];
				frame = (u64 *)frame[0];
				
				if ((u64)frame < 0xffff800000000000)
					break;

				if (!(i & 3))
					printf("\n");
				
				printf("0x%016llx ", stackptr);
			}
		}

		struct TimerInt : public IRQ::Interrupt {
			bool action()
			{
				Time::monotonic_timers->run();
				return true;
			}
		};
		
		TimerInt timer_int;
	}

	using IRQ::i8259;
	::Threads::Thread *init_thread;
		
	void arch_init()
	{
		init_thread = &x64_init_stack.thread;
		Priv::early_adjust_mappings();
		Priv::set_idt();
		Priv::MultiBoot::process_info();
		i8259.init();
		
		u64 tss_addr = reinterpret_cast<u64>(&Priv::tss) + 4;
		x64_gdt[4].base_low = tss_addr & 0xffff;
		x64_gdt[4].base_mid = (tss_addr & 0xff0000) >> 16;
		x64_gdt[4].base_high = (tss_addr & 0xff000000) >> 24;
		
		asm volatile("ltr %w0" : : "r" (0x20) : "memory");
		init_thread->aspace = new Mem::AddrSpace(x64_init_ptbl_l4);
		init_thread->active_aspace = init_thread->aspace;
	}

	void timer_init()
	{
		IRQ::InterruptSlot *timer = i8259.get_slot(0);
		i8259.request_int(timer, &Priv::timer_int);
	}
	
	void ArchThread::init(void *entry, void *arg)
	{
		void **stack = reinterpret_cast<void **>(this);
		
		*--stack = arg;
		*--stack = entry;
		
		rsp = stack;
		rbp = 0;
		jump_to_init = 1;
	}
}

using Arch::Priv::show_regs;

extern "C" void x64_do_diverr(u64 *stack)
{
	printf("Division error at 0x%04llx:0x%016llx\n", stack[17], stack[16]);
	show_regs(stack);
	for(;;);
}

extern "C" void x64_do_invalid_insn(u64 *stack)
{
	printf("Invalid instruction at 0x%04llx:0x%016llx\n", stack[17], stack[16]);
	show_regs(stack);
	for(;;);
}

extern "C" void x64_do_page_fault(u64 *stack, u64 fault_addr, u32 error_code)
{
	Mem::AddrSpace *as;
	
	if (in_fault)
		for(;;);
		
	// A reserved bit was set in the PTE; this is always a bug.
	if (error_code & 8)
		goto bad_fault;
	
	// Don't try to fix up a page fault if interrupts were disabled.  It is an
	// error to access non-locked pages with interrupts disabled.  Trying to
	// fix it up in the case of an access that would be legitimate if interrupts
	// were enabled would simply mask the loss of atomicity, and trying to grab
	// locks to look up the address if it is a completely bad reference won't
	// accomplish much other than decreasing the odds that the fault message
	// gets out.
	
	if (!(stack[18] & 0x200))
		goto bad_fault;
	
	// Don't allow fault-ins using a borrowed addr-space.
	as = curthread->aspace;

	if (!as || curthread == Arch::init_thread)
		goto bad_fault;
	
	ll_ints_on();
	
	// FIXME: no-exec
	if (as->handle_fault(fault_addr, error_code & 2,
	                     false /* error_code & 16 */, error_code & 4))
		return;

	// FIXME: throw exception to user

bad_fault:
	ll_ints_off();
	in_fault++;
	
	printf("Page fault at 0x%04llx:0x%016llx for 0x%016llx\n",
	       stack[17], stack[16], fault_addr);
	printf("Error code: 0x%04x\n", error_code);
	
	show_regs(stack);

	for(;;);
}

extern "C" void x64_do_gpf(u64 *stack, u32 error_code)
{
	if (in_fault)
		for(;;);
	
	in_fault++;

	printf("General protection fault at 0x%04llx:0x%016llx, "
	       "Error code: 0x%04x\n",
	       stack[17], stack[16], error_code);

	show_regs(stack);
	
	for(;;);
}

extern "C" void x64_do_irq(int irq)
{
	IRQ::i8259.handle_irq(irq - 0x20);
}
