// core/irq.cc -- generic IRQ handling
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.


#include <kern/kernel.h>
#include <kern/irq.h>

namespace IRQ {
	void InterruptController::request_int(InterruptSlot *slot, Interrupt *irq)
	{
		AutoSpinLockIRQ autolock(lock);
		
		u32 irqnum = get_irqnum(slot);
		Interrupt *list = slot->first_int;
		Interrupt **insert = &slot->first_int;
		
		while (list) {
			if (list == irq)
				throw_idl(ResourceBusy, -1, countarray<ORBMM>("IRQ in use"));
			
			insert = &list->next;
			list = *insert;
		}
		
		*insert = irq;
		irq->slot = slot;
		irq->controller = this;
		
		if (insert == &slot->first_int && slot->mask_count == 0) {
			// There were no existing handlers, so it's currently masked
			unmask(irqnum);
		}
	}
	
	void InterruptController::free_int(Interrupt *irq)
	{
		InterruptSlot *slot = irq->slot;
		u32 irqnum = get_irqnum(slot);
	
		DroppableAutoSpinLockIRQ autolock(lock);
		
		Interrupt *list = slot->first_int;
		Interrupt **remove = &slot->first_int;

		while (list) {
			if (list == irq) {
				*remove = list->next;

				if (remove == &slot->first_int && !list->next &&
				    slot->mask_count == 0)
				{
					// It was the last handler for this IRQ num.
					mask(irqnum);
				}

				autolock.unlock();
				wait_for_irq(slot);
				return;
			}
			
			remove = &list->next;
			list = *remove;
		}
		
		throw_idl(InvalidReference, 0, nullarray);
	}
	
	bool InterruptController::handle_irq(int irq)
	{
		DroppableAutoSpinLock autolock(lock);
		bool handled_one = false;
		bool irq_specified = irq >= 0;
	
		in_irq = true;

		do {
			if (!irq_specified) {
				irq = get_pending_irq();
				
				if (irq < 0)
					break;
			}
			
			mask_and_ack(irq);

			InterruptSlot *slot = get_slot(irq);
			
			if (slot->mask_count > 0 || slot->flags & InterruptSlot::Running) {
				slot->flags |= InterruptSlot::Pending;
				continue;
			}
				
			slot->flags |= InterruptSlot::Running;
			assert(!(slot->flags & InterruptSlot::Pending));
		
			autolock.unlock();

			for (Interrupt *i = slot->first_int; i; i = i->next)
				handled_one |= i->action();

			autolock.lock();

			slot->flags &= ~InterruptSlot::Running;
			
			if (slot->mask_count == 0)
				unmask(irq);
		} while (!irq_specified);

		in_irq = false;
		return handled_one;
	}
	
	void InterruptController::rec_mask_nowait(InterruptSlot *slot)
	{
		AutoSpinLockIRQ autolock(lock);
		u32 irq = get_irqnum(slot);
		
		if (slot->mask_count++ == 0)
			mask(irq);
	}
	
	void InterruptController::rec_mask(InterruptSlot *slot)
	{
		rec_mask_nowait(slot);
		wait_for_irq(slot);
	}

	void InterruptController::rec_unmask(InterruptSlot *slot)
	{
		bool was_pending = false;
		DroppableAutoSpinLockRecIRQ autolock(lock);
		
		if (--slot->mask_count == 0) {
			unmask(get_irqnum(slot));
			
			if (slot->flags & InterruptSlot::Pending) {
				was_pending = true;
				slot->flags &= ~InterruptSlot::Pending;
				slot->flags |= InterruptSlot::Running;
			}
		}

		autolock.unlock();

		if (was_pending) {
			// Only really necessary for edge-triggered interrupts
			
			for (Interrupt *i = slot->first_int; i; i = i->next)
				i->action();
			
			autolock.lock();
			slot->flags &= ~InterruptSlot::Running;
			autolock.unlock();
		}
	}
	
	void InterruptController::wait_for_irq(InterruptSlot *slot)
	{
		while (slot->flags & InterruptSlot::Running)
			ll_busywait();
	}
	
	bool in_irq;
}

//#include "irq-server/footer.cc"
