// mem/pagealloc.cc -- low-level page allocator
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/kernel.h>
#include <kern/pagealloc.h>
#include <kern/paging.h>
#include <kern/bitops.h>
#include <kern/list.h>
#include <kern/mem.h>
#include <limits.h>

namespace Mem {
	Page *pages, *last_page;
	inline size_t PageAllocZone::chunk_size(Page *start)
	{
		ptrdiff_t ret = start->free.otherside - start + 1;
		assert(ret >= 1);
		assert(static_cast<size_t>(ret) <= zonesize);
		return static_cast<size_t>(ret);
	}

	inline Page *PageAllocZone::bin_to_head(int bin)
	{
		return &bins[bin];
	}

	inline uint PageAllocZone::bin_to_size(int bin)
	{
		return 1 << bin;
	}
	
	inline int PageAllocZone::size_to_bin_alloc(size_t size)
	{
		int bin = ll_get_order_round_up(size);
		
		if (bin >= num_bins)
			bin = num_bins - 1;
		
		return bin;
	}

	inline int PageAllocZone::size_to_bin_free(size_t size)
	{
		int bin = ll_get_order_round_down(size);
		
		if (bin >= num_bins)
			bin = num_bins - 1;
		
		return bin;
	}

	inline void PageAllocZone::remove_chunk(Page *chunk, int bin)
	{
		chunk->chunk_rmap_list.del();
		
		// If bin's list is empty, forward it to the next non-empty bin.
		
		Page *head = bin_to_head(bin);
		if (head->chunk_rmap_list.empty()) {
			Page *newforward = bin_to_head(bin + 1)->free.otherside;
			assert(head->free.otherside == head);

			for (Page *p = head;
			     p - bins >= 0 && p->free.otherside == head; p--)
				p->free.otherside = newforward;
		}
	}

	inline void PageAllocZone::add_to_bin(Page *chunk, int bin)
	{
		Page *head = bin_to_head(bin);
		Page *oldforward = head->free.otherside;

		head->chunk_rmap_list.add_front(&chunk->chunk_rmap_list);
		
		// Remove bin's forwarding if it was empty, along with smaller
		// bins pointing past this bin.
		
		if (oldforward != head) {
			Page *p = head;
			do {
				p->free.otherside = head;
				p--;
			} while (p - bins >= 0 && p->free.otherside == oldforward);
		}
	}

	// Break off a piece from the end of the chunk, rather than the start. 
	// This way, only the size of the remaining chunk needs to be adjusted
	// as long as it still fits in the same bin.
	
	Page *PageAllocZone::shrink_chunk(Page *start, int num_pages,
	                                  size_t chunk_size, int bin)
	{
		size_t size_left = chunk_size - num_pages;
		Page *newend = start + size_left - 1;
		
		start->free.otherside = newend;
		newend->free.otherside = start;
		
		if (size_left < bin_to_size(bin)) {
			remove_chunk(start, bin);
			add_to_bin(start, size_to_bin_alloc(size_left));
		}
		
		return newend + 1;
	}
	
	struct Page *PageAllocZone::alloc(uint num_pages)
	{
		Lock::AutoSpinLockRecIRQ autolock(lock);
	
		assert(num_pages > 0);
		assert(num_pages <= bin_to_size(num_bins - 1));
	
		int bin = size_to_bin_alloc(num_pages);
		Page *head = bin_to_head(bin)->free.otherside;
		int realbin = head - bins;
		Util::List *list = &head->chunk_rmap_list;
		
		if (list->empty()) {
			assert(realbin == num_bins - 1);
			return NULL;
		}
		
		head = list->next->listentry(Page, chunk_rmap_list);
		
		assert(head->flags & Page::Free);
		
		size_t size = chunk_size(head);
		assert(size >= bin_to_size(bin));
		assert(size >= (uint)num_pages);
		
		if (size != num_pages)
			head = shrink_chunk(head, num_pages, size, realbin);
		else
			remove_chunk(head, realbin);

		for (Page *p = head; p != head + num_pages; p++) {
			assert(p->flags & Page::Free);
			assert(!(p->flags & Page::InUse));
			assert(p->chunk_rmap_list.empty());
			p->flags = (p->flags & ~Page::Free) | Page::InUse;
			p->inuse.refcount = 1;
		}

		return head;
	}
	
	void PageAllocZone::free(struct Page *head, size_t num_pages)
	{
		Lock::AutoSpinLockRecIRQ autolock(lock);

		assert(num_pages > 0);
	
		for (Page *p = head; p != head + num_pages; p++) {
			assert(!(p->flags & Page::Free));
			assert(p->chunk_rmap_list.empty());
			p->flags = (p->flags & ~Page::InUse) | Page::Free;
		}
		
		// Combine the newly free chunk with any adjacent free chunks,
		// regardless of what bin they are in.

		Page *prevpage = head - 1;
		Page *nextpage = head + num_pages;
		
		int bin = -1;
				
		if (prevpage - pages >= start && (prevpage->flags & Page::Free)) {
			Page *prevchunk = prevpage->free.otherside;
			assert(prevchunk->flags & Page::Free);

			Page *end = head + num_pages - 1;
			prevchunk->free.otherside = end;
			end->free.otherside = prevchunk;
			
			size_t prevsize = head - prevchunk;
			assert(prevsize > 0);
			assert(prevsize < zonesize);

			head = prevchunk;
			num_pages += prevsize;
			bin = size_to_bin_free(prevsize);
		}

		if (nextpage - pages <= end && nextpage->flags & Page::Free) {
			size_t prevsize = chunk_size(nextpage);
			num_pages += prevsize;
			Page *end = nextpage->free.otherside;
			
			remove_chunk(nextpage, size_to_bin_free(prevsize));
			
			end->free.otherside = head;
			head->free.otherside = end;
		}
		
		int newbin = size_to_bin_free(num_pages);
		
		if (bin != newbin) {
			if (bin != -1) {
				remove_chunk(head, bin);
				assert(head->free.otherside == head + num_pages - 1);
			}

			head->free.otherside = head + num_pages - 1;
			add_to_bin(head, newbin);
		}
	}
	
	void PageAllocZone::init(uintptr_t base, size_t size)
	{
		assert(size > 0);
		assert(base + size <= Arch::mem_end / Arch::page_size);
	
		zonesize = size;
		start = base;
		end = base + size - 1;

		for (Page *p = &pages[start]; p <= &pages[end]; p++) {
			assert(p->flags == 0);
			new(p) Page;
			p->zone = this;
		}
		
		for (int i = 0; i < num_bins; i++) {
			Page *bin = bin_to_head(i);
			
			bin->free.otherside = bin_to_head(num_bins - 1);
			bin->chunk_rmap_list.init();
		}
	}
	
	Page *PageAlloc::alloc(uint num_pages, PageAllocZone *const *zonelist)
	{
		while (*zonelist) {
			Page *ret = (*zonelist)->alloc(num_pages);
			if (ret)
				return ret;
			
			zonelist++;
		}
		
		// FIXME: Deliver System.Traps.ReduceMemoryUsage first
		throw_idl(OutOfMemory);
	}
	
	void Page::free_page()
	{
		PageAlloc::free(this, 1);
	}

	PageAlloc page_alloc;
}
